"""Universal LLM client with factory pattern for provider selection."""

from typing import Dict, Optional, AsyncIterator
from .models import (
    CompletionRequest,
    CompletionResponse,
    ModelCapabilities,
    Message,
    MessageRole,
    ProviderType,
)
from .providers import (
    BaseLLMProvider,
    OpenAIProvider,
    AnthropicProvider,
    DeepseekProvider,
    MistralProvider,
)
from .exceptions import ProviderError, ModelNotSupportedError


class UniversalLLMClient:
    """Universal client for interacting with different LLM providers."""

    _provider_classes = {
        ProviderType.OPENAI: OpenAIProvider,
        ProviderType.ANTHROPIC: AnthropicProvider,
        ProviderType.DEEPSEEK: DeepseekProvider,
        ProviderType.MISTRAL: MistralProvider,
    }

    def __init__(self, provider: Optional[ProviderType] = None, **kwargs) -> None:
        """Initialize the universal client.

        Args:
            provider: The provider to use (if not specified, will auto-detect)
            **kwargs: Provider-specific configuration
        """
        self.provider_instance: Optional[BaseLLMProvider] = None
        self.provider_type = provider
        self.config = kwargs

        if provider:
            self._initialize_provider(provider, **kwargs)

    def _initialize_provider(self, provider: ProviderType, **kwargs) -> None:
        """Initialize a specific provider.

        Args:
            provider: The provider type to initialize
            **kwargs: Provider-specific configuration
        """
        if provider not in self._provider_classes:
            raise ProviderError(f"Unsupported provider: {provider}")

        provider_class = self._provider_classes[provider]
        self.provider_instance = provider_class(**kwargs)
        self.provider_type = provider

    def _auto_detect_provider(self, model: str) -> ProviderType:
        """Auto-detect provider based on model name without instantiation."""
        # First, try class-level supported model checks
        for provider_type, provider_class in self._provider_classes.items():
            try:
                if provider_class.supports_model(model):  # type: ignore[attr-defined]
                    return provider_type
            except Exception:
                continue

        # Fallback heuristics
        model_lower = model.lower()

        if any(keyword in model_lower for keyword in ["gpt", "openai"]):
            return ProviderType.OPENAI
        if any(keyword in model_lower for keyword in ["claude", "anthropic"]):
            return ProviderType.ANTHROPIC
        if "deepseek" in model_lower:
            return ProviderType.DEEPSEEK
        if any(keyword in model_lower for keyword in ["mistral", "mixtral"]):
            return ProviderType.MISTRAL

        raise ModelNotSupportedError(
            f"Could not auto-detect provider for model: {model}"
        )

    def set_provider(self, provider: ProviderType, **kwargs) -> None:
        """Set or change the provider.

        Args:
            provider: The provider type to use
            **kwargs: Provider-specific configuration
        """
        self._initialize_provider(provider, **kwargs)

    def get_supported_models(
        self, provider: Optional[ProviderType] = None
    ) -> Dict[ProviderType, list]:
        """Get supported models for all or specific provider using class-level data.

        Args:
            provider: Specific provider to get models for (if None, gets all)

        Returns:
            Dictionary mapping provider types to their supported models
        """
        if provider:
            if provider not in self._provider_classes:
                raise ProviderError(f"Unsupported provider: {provider}")
            provider_class = self._provider_classes[provider]
            return {provider: list(getattr(provider_class, "SUPPORTED_MODELS", []))}

        all_models: Dict[ProviderType, list] = {}
        for provider_type, provider_class in self._provider_classes.items():
            all_models[provider_type] = list(
                getattr(provider_class, "SUPPORTED_MODELS", [])
            )

        return all_models

    def get_model_capabilities(
        self, model: str, provider: Optional[ProviderType] = None
    ) -> ModelCapabilities:
        """Get capabilities for a specific model.

        Args:
            model: Model identifier
            provider: Provider to use (if not specified, will auto-detect)

        Returns:
            Model capabilities
        """
        if not provider:
            provider = self._auto_detect_provider(model)

        if not self.provider_instance or self.provider_type != provider:
            self._initialize_provider(provider, **self.config)

        return self.provider_instance.get_model_capabilities(model)

    async def complete(
        self,
        messages: list,
        model: str,
        provider: Optional[ProviderType] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        **kwargs,
    ) -> CompletionResponse:
        """Generate a completion.

        Args:
            messages: List of messages (can be strings or Message objects)
            model: Model identifier
            provider: Provider to use (if not specified, will auto-detect)
            max_tokens: Maximum tokens to generate
            temperature: Sampling temperature
            top_p: Top-p sampling parameter
            **kwargs: Additional provider-specific parameters

        Returns:
            Completion response
        """
        if not provider:
            provider = self._auto_detect_provider(model)

        if not self.provider_instance or self.provider_type != provider:
            self._initialize_provider(provider, **self.config)

        # Convert messages to Message objects if they're strings
        processed_messages = []
        for i, msg in enumerate(messages):
            if isinstance(msg, str):
                # First message is system if only one, otherwise alternate user/assistant
                if i == 0 and len(messages) == 1:
                    role = MessageRole.USER
                elif i == 0:
                    role = MessageRole.SYSTEM
                elif i % 2 == 1:
                    role = MessageRole.USER
                else:
                    role = MessageRole.ASSISTANT
                processed_messages.append(Message(role=role, content=msg))
            elif isinstance(msg, dict):
                processed_messages.append(
                    Message(role=MessageRole(msg["role"]), content=msg["content"])
                )
            else:
                processed_messages.append(msg)

        request = CompletionRequest(
            messages=processed_messages,
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            extra_params=kwargs,
        )

        return await self.provider_instance.complete(request)

    async def stream_complete(
        self,
        messages: list,
        model: str,
        provider: Optional[ProviderType] = None,
        max_tokens: Optional[int] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        **kwargs,
    ) -> AsyncIterator[str]:
        """Generate a streaming completion.

        Args:
            messages: List of messages (can be strings or Message objects)
            model: Model identifier
            provider: Provider to use (if not specified, will auto-detect)
            max_tokens: Maximum tokens to generate
            temperature: Sampling temperature
            top_p: Top-p sampling parameter
            **kwargs: Additional provider-specific parameters

        Yields:
            Chunks of the completion
        """
        if not provider:
            provider = self._auto_detect_provider(model)

        if not self.provider_instance or self.provider_type != provider:
            self._initialize_provider(provider, **self.config)

        # Convert messages to Message objects if they're strings
        processed_messages = []
        for i, msg in enumerate(messages):
            if isinstance(msg, str):
                # First message is system if only one, otherwise alternate user/assistant
                if i == 0 and len(messages) == 1:
                    role = MessageRole.USER
                elif i == 0:
                    role = MessageRole.SYSTEM
                elif i % 2 == 1:
                    role = MessageRole.USER
                else:
                    role = MessageRole.ASSISTANT
                processed_messages.append(Message(role=role, content=msg))
            elif isinstance(msg, dict):
                processed_messages.append(
                    Message(role=MessageRole(msg["role"]), content=msg["content"])
                )
            else:
                processed_messages.append(msg)

        request = CompletionRequest(
            messages=processed_messages,
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            stream=True,
            extra_params=kwargs,
        )

        async for chunk in self.provider_instance.stream_complete(request):
            yield chunk



"""Custom exceptions for univllm package."""


class UniversalLLMError(Exception):
    """Base exception for univllm package."""

    pass


class ProviderError(UniversalLLMError):
    """Exception raised when there's an error with a provider."""

    pass


class ModelNotSupportedError(UniversalLLMError):
    """Exception raised when a model is not supported by a provider."""

    pass


class ConfigurationError(UniversalLLMError):
    """Exception raised when there's a configuration issue."""

    pass


class AuthenticationError(UniversalLLMError):
    """Exception raised when authentication fails."""

    pass



"""Basic tests for univllm package."""

import pytest
from univllm import UniversalLLMClient
from univllm.models import ModelCapabilities, MessageRole, Message, ProviderType
from univllm.exceptions import ModelNotSupportedError


def test_imports():
    """Test that all main classes can be imported."""
    from univllm import (
        UniversalLLMClient,
        ProviderType,
        UniversalLLMError,
        ProviderError,
        ModelNotSupportedError,
    )

    assert UniversalLLMClient is not None
    assert ProviderType is not None


def test_client_creation():
    """Test that client can be created without errors."""
    client = UniversalLLMClient()
    assert client is not None


def test_auto_detection():
    """Test model provider auto-detection."""
    client = UniversalLLMClient()

    # Test OpenAI models
    assert client._auto_detect_provider("gpt-3.5-turbo") == ProviderType.OPENAI
    assert client._auto_detect_provider("gpt-4") == ProviderType.OPENAI

    # Test Anthropic models
    assert (
        client._auto_detect_provider("claude-3-sonnet-20240229")
        == ProviderType.ANTHROPIC
    )
    assert client._auto_detect_provider("claude-2.1") == ProviderType.ANTHROPIC

    # Test Deepseek models
    assert client._auto_detect_provider("deepseek-chat") == ProviderType.DEEPSEEK
    assert client._auto_detect_provider("deepseek-coder") == ProviderType.DEEPSEEK

    # Test Mistral models
    assert client._auto_detect_provider("mistral-large-latest") == ProviderType.MISTRAL
    assert client._auto_detect_provider("mixtral-8x7b") == ProviderType.MISTRAL


def test_unsupported_model():
    """Test handling of unsupported models."""
    client = UniversalLLMClient()

    with pytest.raises(ModelNotSupportedError):
        client._auto_detect_provider("nonexistent-model")


def test_provider_types():
    """Test provider type enum values."""
    assert ProviderType.OPENAI == "openai"
    assert ProviderType.ANTHROPIC == "anthropic"
    assert ProviderType.DEEPSEEK == "deepseek"
    assert ProviderType.MISTRAL == "mistral"


def test_message_creation():
    """Test message model creation."""
    msg = Message(role=MessageRole.USER, content="Hello")
    assert msg.role == MessageRole.USER
    assert msg.content == "Hello"


def test_model_capabilities():
    """Test model capabilities structure."""
    caps = ModelCapabilities(
        supports_system_messages=True,
        supports_function_calling=False,
        supports_vision=True,
        context_window=4096,
    )
    assert caps.supports_system_messages is True
    assert caps.supports_function_calling is False
    assert caps.supports_vision is True
    assert caps.context_window == 4096


def test_supported_models_structure():
    """Test that get_supported_models returns proper structure."""
    client = UniversalLLMClient()
    models = client.get_supported_models()

    # Should return a dict with all provider types
    assert isinstance(models, dict)
    assert ProviderType.OPENAI in models
    assert ProviderType.ANTHROPIC in models
    assert ProviderType.DEEPSEEK in models
    assert ProviderType.MISTRAL in models

    # Each value should be a list
    for provider, model_list in models.items():
        assert isinstance(model_list, list)


if __name__ == "__main__":
    pytest.main([__file__])



"""Deepseek provider implementation."""

import json
import os
from typing import List, Optional, AsyncIterator
import httpx

from ..supported_models import DEEPSEEK_SUPPORTED_MODELS
from ..models import (
    CompletionRequest,
    CompletionResponse,
    ModelCapabilities,
    ProviderType,
)
from ..exceptions import ProviderError, ModelNotSupportedError, AuthenticationError
from .base import BaseLLMProvider


class DeepseekProvider(BaseLLMProvider):
    """Deepseek provider for Deepseek models."""

    SUPPORTED_MODELS: List[str] = DEEPSEEK_SUPPORTED_MODELS

    def __init__(
        self, api_key: Optional[str] = None, base_url: Optional[str] = None, **kwargs
    ) -> None:
        """Initialize Deepseek provider.

        Args:
            api_key: Deepseek API key (if not provided, will use DEEPSEEK_API_KEY env var)
            base_url: Base URL for Deepseek API (default: https://api.deepseek.com)
            **kwargs: Additional configuration
        """
        api_key = api_key or os.getenv("DEEPSEEK_API_KEY")
        if not api_key:
            raise AuthenticationError("Deepseek API key is required")

        super().__init__(api_key=api_key, **kwargs)
        self.base_url = base_url or "https://api.deepseek.com"
        self.client = httpx.AsyncClient()

    @property
    def provider_type(self) -> ProviderType:
        """Return the provider type."""
        return ProviderType.DEEPSEEK

    def get_model_capabilities(self, model: str) -> ModelCapabilities:
        """Get capabilities for a specific Deepseek model."""
        if not self.validate_model(model):
            raise ModelNotSupportedError(
                f"Model {model} is not supported by Deepseek provider"
            )

        # Default capabilities for Deepseek models
        capabilities = ModelCapabilities(
            supports_system_messages=True,
            supports_function_calling=False,
            supports_streaming=True,
            supports_vision=False,
            context_window=32000,
            max_tokens=4096,
        )

        return capabilities

    async def complete(self, request: CompletionRequest) -> CompletionResponse:
        """Generate a completion using Deepseek."""
        if not self.validate_model(request.model):
            raise ModelNotSupportedError(
                f"Model {request.model} is not supported by Deepseek provider"
            )

        try:
            # Prepare the request data
            data = self.prepare_request(request)

            headers = {
                "Authorization": f"Bearer {self.api_key}",
                "Content-Type": "application/json",
            }

            # Make the API call
            response = await self.client.post(
                f"{self.base_url}/v1/chat/completions", json=data, headers=headers
            )
            response.raise_for_status()

            result = response.json()

            # Extract the response
            content = (
                result["choices"][0]["message"]["content"]
                if result.get("choices")
                else ""
            )
            usage = result.get("usage", {})

            return CompletionResponse(
                content=content,
                model=result.get("model", request.model),
                usage={
                    "prompt_tokens": usage.get("prompt_tokens", 0),
                    "completion_tokens": usage.get("completion_tokens", 0),
                    "total_tokens": usage.get("total_tokens", 0),
                }
                if usage
                else None,
                finish_reason=result["choices"][0].get("finish_reason")
                if result.get("choices")
                else None,
                provider=self.provider_type,
            )

        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                raise AuthenticationError(f"Deepseek authentication failed: {e}")
            elif e.response.status_code == 429:
                raise ProviderError(f"Deepseek rate limit exceeded: {e}")
            else:
                raise ProviderError(f"Deepseek API error: {e}")
        except Exception as e:
            raise ProviderError(f"Deepseek provider error: {e}")

    async def stream_complete(self, request: CompletionRequest) -> AsyncIterator[str]:
        """Generate a streaming completion using Deepseek."""
        if not self.validate_model(request.model):
            raise ModelNotSupportedError(
                f"Model {request.model} is not supported by Deepseek provider"
            )

        try:
            # Prepare the request data with streaming enabled
            data = self.prepare_request(request)
            data["stream"] = True

            headers = {
                "Authorization": f"Bearer {self.api_key}",
                "Content-Type": "application/json",
            }

            # Make the streaming API call
            async with self.client.stream(
                "POST",
                f"{self.base_url}/v1/chat/completions",
                json=data,
                headers=headers,
            ) as response:
                response.raise_for_status()

                async for line in response.aiter_lines():
                    if line.startswith("data: "):
                        chunk_data = line[6:]  # Remove "data: " prefix
                        if chunk_data.strip() == "[DONE]":
                            break

                        try:
                            chunk = json.loads(chunk_data)
                            if chunk.get("choices") and chunk["choices"][0].get(
                                "delta", {}
                            ).get("content"):
                                yield chunk["choices"][0]["delta"]["content"]
                        except json.JSONDecodeError:
                            continue

        except httpx.HTTPStatusError as e:
            if e.response.status_code == 401:
                raise AuthenticationError(f"Deepseek authentication failed: {e}")
            elif e.response.status_code == 429:
                raise ProviderError(f"Deepseek rate limit exceeded: {e}")
            else:
                raise ProviderError(f"Deepseek API error: {e}")
        except Exception as e:
            raise ProviderError(f"Deepseek provider error: {e}")

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.client.aclose()



"""Provider implementations for different LLM services."""

from .base import BaseLLMProvider
from .openai_provider import OpenAIProvider
from .anthropic_provider import AnthropicProvider
from .deepseek_provider import DeepseekProvider
from .mistral_provider import MistralProvider
from ..models import ProviderType

__all__ = [
    "BaseLLMProvider",
    "OpenAIProvider",
    "AnthropicProvider",
    "DeepseekProvider",
    "MistralProvider",
]



"""OpenAI provider implementation."""

import os
from typing import List, Optional, AsyncIterator
import openai

from ..supported_models import OPENAI_SUPPORTED_MODELS
from ..models import (
    CompletionRequest,
    CompletionResponse,
    ModelCapabilities,
    ProviderType,
)
from ..exceptions import ProviderError, ModelNotSupportedError, AuthenticationError
from .base import BaseLLMProvider


class OpenAIProvider(BaseLLMProvider):
    """OpenAI provider for GPT models."""

    # Class-level supported models (prefixes or exact names)
    SUPPORTED_MODELS: List[str] = OPENAI_SUPPORTED_MODELS

    def __init__(self, api_key: Optional[str] = None, **kwargs) -> None:
        """Initialize OpenAI provider.

        Args:
            api_key: OpenAI API key (if not provided, will use OPENAI_API_KEY env var)
            **kwargs: Additional configuration
        """
        api_key = api_key or os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise AuthenticationError("OpenAI API key is required")

        super().__init__(api_key=api_key, **kwargs)
        self.client = openai.AsyncOpenAI(api_key=api_key)

    @property
    def provider_type(self) -> ProviderType:
        """Return the provider type."""
        return ProviderType.OPENAI

    def get_model_capabilities(self, model: str) -> ModelCapabilities:
        """Get capabilities for a specific OpenAI model."""
        if not self.validate_model(model):
            raise ModelNotSupportedError(
                f"Model {model} is not supported by OpenAI provider"
            )

        # Default capabilities for OpenAI models
        capabilities = ModelCapabilities(
            supports_system_messages=True,
            supports_function_calling=True,
            supports_streaming=True,
        )

        # Model-specific capabilities
        if "gpt-4" in model:
            capabilities.context_window = 8192 if "turbo" not in model else 128000
            capabilities.max_tokens = 4096
            if "vision" in model:
                capabilities.supports_vision = True
        elif "gpt-3.5-turbo" in model:
            capabilities.context_window = 4096
            capabilities.max_tokens = 4096

        return capabilities

    async def complete(self, request: CompletionRequest) -> CompletionResponse:
        """Generate a completion using OpenAI."""
        if not self.validate_model(request.model):
            raise ModelNotSupportedError(
                f"Model {request.model} is not supported by OpenAI provider"
            )

        try:
            # Prepare the request data
            data = self.prepare_request(request)

            # Make the API call
            response = await self.client.chat.completions.create(**data)

            # Extract the response
            content = response.choices[0].message.content or ""
            usage = (
                {
                    "prompt_tokens": response.usage.prompt_tokens
                    if response.usage
                    else 0,
                    "completion_tokens": response.usage.completion_tokens
                    if response.usage
                    else 0,
                    "total_tokens": response.usage.total_tokens
                    if response.usage
                    else 0,
                }
                if response.usage
                else None
            )

            return CompletionResponse(
                content=content,
                model=response.model,
                usage=usage,
                finish_reason=response.choices[0].finish_reason,
                provider=self.provider_type,
            )

        except openai.AuthenticationError as e:
            raise AuthenticationError(f"OpenAI authentication failed: {e}")
        except openai.RateLimitError as e:
            raise ProviderError(f"OpenAI rate limit exceeded: {e}")
        except openai.APIError as e:
            raise ProviderError(f"OpenAI API error: {e}")
        except Exception as e:
            raise ProviderError(f"OpenAI provider error: {e}")

    async def stream_complete(self, request: CompletionRequest) -> AsyncIterator[str]:
        """Generate a streaming completion using OpenAI."""
        if not self.validate_model(request.model):
            raise ModelNotSupportedError(
                f"Model {request.model} is not supported by OpenAI provider"
            )

        try:
            # Prepare the request data with streaming enabled
            data = self.prepare_request(request)
            data["stream"] = True

            # Make the streaming API call
            stream = await self.client.chat.completions.create(**data)

            async for chunk in stream:
                if chunk.choices and chunk.choices[0].delta.content:
                    yield chunk.choices[0].delta.content

        except openai.AuthenticationError as e:
            raise AuthenticationError(f"OpenAI authentication failed: {e}")
        except openai.RateLimitError as e:
            raise ProviderError(f"OpenAI rate limit exceeded: {e}")
        except openai.APIError as e:
            raise ProviderError(f"OpenAI API error: {e}")
        except Exception as e:
            raise ProviderError(f"OpenAI provider error: {e}")
